// @ts-nocheck

import { io } from "socket.io-client";

import {
  Configuration,
  ConfigurationParameters,
  ConnectionsApi,
  CredentialsApi,
  CredentialsOffersApi,
  CredentialsProofApi,
  SchemaApi,
} from './swaggerGateway';
import EventEmitter from "events";
import TypedEmitter from "typed-emitter";

export interface Config {
  websocketUrl: string;
  gatewayUrl: string;
  getToken: () => Promise<string>;
}

let config: Config | null = null;

export const setConfig = (cfg: Config) => config = cfg;

export type GatewayEvents =  {
  "message": (message: GatewayMessageType) => void;
};

export interface GatewayMessageType {
  id: string;
  specversion: string;
  datacontenttype: string;
  subject: string;
  source: string;
  type: string;
  data: any;
}

class GatewayApi {
  public connections: ConnectionsApi;
  public credentials: CredentialsApi;
  public credentialsOffers: CredentialsOffersApi;
  public credentialsProof: CredentialsProofApi;
  public schema: SchemaApi;

  public events: TypedEmitter<GatewayEvents> = new EventEmitter();

  constructor() {

  }

  public initialize = async () => {
    const cfg = new Configuration();

    const fetchPatched = async (url: string, init?: any) => {
      init = Object.assign({}, init);
      Object.assign(init.headers, {
        Authorization: `Bearer ${await config.getToken()}`
      })

      const conf: any = {};
      let eventId = "-1";
      const events = {};
      conf.fn = (message: GatewayMessageType) => {
        console.log(new Date().getTime() + ' Received message: ', eventId, message);
        if (message.id === eventId) {
          // unsubscribe from events
          console.log('Unsubscribe ' + eventId, conf.fn);
          this.events.off('message', conf.fn);
          clearTimeout(conf.timeoutId);
          conf.resolve({ status: 200, json: () => message.data });
        } else {
          events[message.id] = message;
        }
      };
      this.events.on('message', conf.fn);

      return new Promise(async (resolve, reject) => {
        const TIMEOUT = 5 * 60 * 1000; // 5 minutes;

        conf.resolve = resolve;
        conf.timeoutId = setTimeout(() => {
          console.log('Unsubscribe ' + eventId, conf.fn);
          this.events.off('message', conf.fn);
          reject(new Error('Timeout exception'));
        }, TIMEOUT);

        const gatewayResponse = await fetch(url, init);
        const gatewayData = await gatewayResponse.json();
        eventId = gatewayData.id;
        console.log('Sent request. Event id is: ' + eventId);
        if (events[eventId]) {

          console.log('Unsubscribe ' + eventId, conf.fn);
          this.events.off('message', conf.fn);
          clearTimeout(conf.timeoutId);
          conf.resolve({ status: 200, json: () => events[eventId].data });
        }
      });
    };

    const baseUrl = config.gatewayUrl;

    this.connections = new ConnectionsApi(cfg, baseUrl, fetchPatched);
    this.credentials = new CredentialsApi(cfg, baseUrl, fetchPatched);
    this.credentialsOffers = new CredentialsOffersApi(cfg, baseUrl, fetchPatched);
    this.credentialsProof = new CredentialsProofApi(cfg, baseUrl, fetchPatched);
    this.schema = new SchemaApi(cfg, baseUrl, fetchPatched);
  }

  public async openConnection() {
    const socket = io(`${config.websocketUrl}/`, {
      reconnectionDelayMax: 10000,
      auth: {
        token: await config.getToken()
      },
    });

    socket.on("message", (message) => {
      this.events.emit('message', message);
    })
  }
}

export default new GatewayApi();
