import { Injectable } from "@nestjs/common";
import { AskarService } from "./askar.service";
import {
  AcceptInvitationResponseDto,
  ConnectionNotFoundError,
  CreateCredentialDefinitionRequsetDto,
  CreateCredentialDefinitionResponseDto,
  CreateInvitationResponseDto,
  CreateSchemaRequestDto,
  CreateSchemaResponseDto,
  CredentialNotCreatedError,
  IssueCredentialRequestDto,
  IssueCredentialResponseDto,
  IssueProofRequestDto,
  IssueProofResponseDto,
  MakeBasicMessageRequestDto,
  MakeBasicMessageResponseDto,
  SchemaNotCreatedError,
} from "@ocm-engine/dtos";
import {
  CredentialState,
  ProofExchangeRecord,
  ProofState,
} from "@aries-framework/core";
import { AnonCredsRequestedAttribute } from "@aries-framework/anoncreds";

@Injectable()
export class AgentService {
  constructor(private readonly askar: AskarService) {}

  createInvitation = async () => {
    const outOfBoundRecord = await this.askar.agent.oob.createInvitation();

    const i = new CreateInvitationResponseDto();
    i.invitationUrl = outOfBoundRecord.outOfBandInvitation.toUrl({
      domain: this.askar.agentConfig.agentHttpPeerAddress || this.askar.agentConfig.agentWsPeerAddress || '',
    });

    return i;
  };

  acceptInvitation = async (invitationUrl: string) => {
    const { connectionRecord } =
      await this.askar.agent.oob.receiveInvitationFromUrl(invitationUrl);

    if (typeof connectionRecord === "undefined") {
      throw new ConnectionNotFoundError();
    }

    const r = new AcceptInvitationResponseDto();
    r.connectionName = connectionRecord.theirLabel;
    r.connectionId = connectionRecord.id;
    r.did = connectionRecord.did;
    r.invitationDid = connectionRecord.invitationDid;
    r.outOfBandId = connectionRecord.outOfBandId;
    r.createdAt = connectionRecord.createdAt;

    return r;
  };

  fetchConnections() {
    return this.askar.agent.connections.getAll();
  }

  getConnectionById = (id: string) => {
    return this.askar.agent.connections.findById(id);
  };

  fetchSchemas = () => {
    return this.askar.agent.modules.anoncreds.getCreatedSchemas({});
  };

  getSchemaById = (schemaId: string) => {
    return this.askar.agent.modules.anoncreds.getSchema(schemaId);
  };

  createSchema = async (schema: CreateSchemaRequestDto) => {
    const dids = await this.askar.agent.dids.getCreatedDids({ method: "indy" });

    const schemaResult =
      await this.askar.agent.modules.anoncreds.registerSchema({
        schema: {
          name: schema.name,
          issuerId: dids[0].did,
          attrNames: schema.attributes,
          version: schema.version,
        },
        options: {},
      });

    if (schemaResult.schemaState.state !== "finished") {
      throw new SchemaNotCreatedError();
    }

    const response = new CreateSchemaResponseDto();

    response.name = schemaResult.schemaState.schema.name;
    response.schemaId = schemaResult.schemaState.schemaId;
    response.issuerId = schemaResult.schemaState.schema.issuerId;
    response.version = schemaResult.schemaState.schema.version;
    response.attributes = schemaResult.schemaState.schema.attrNames;

    return response;
  };

  createCredentialDefinition = async (
    credentialDefinitionDto: CreateCredentialDefinitionRequsetDto,
  ) => {
    const dids = await this.askar.agent.dids.getCreatedDids({ method: "indy" });

    const credDef =
      await this.askar.agent.modules.anoncreds.registerCredentialDefinition({
        credentialDefinition: {
          tag: credentialDefinitionDto.tag,
          issuerId: dids[0].did,
          schemaId: credentialDefinitionDto.schemaId,
        },
        options: {},
      });

    if (credDef.credentialDefinitionState.state !== "finished") {
      throw new CredentialNotCreatedError();
    }

    const response = new CreateCredentialDefinitionResponseDto();
    response.credentialDefinitionId =
      credDef.credentialDefinitionState.credentialDefinitionId;
    response.schemaId =
      credDef.credentialDefinitionState.credentialDefinition.schemaId;
    response.isserId =
      credDef.credentialDefinitionState.credentialDefinition.issuerId;
    response.tag = credDef.credentialDefinitionState.credentialDefinition.tag;

    return response;
  };

  issueCredential = async (issueCredentialDto: IssueCredentialRequestDto) => {
    console.log("tuk sym", JSON.stringify(issueCredentialDto, null, 2));
    try {
      const credentialExchangeRecord =
        await this.askar.agent.credentials.offerCredential({
          protocolVersion: "v2",
          connectionId: issueCredentialDto.connectionId,
          credentialFormats: {
            anoncreds: {
              credentialDefinitionId: issueCredentialDto.credentialDefinitionId,
              attributes: issueCredentialDto.attributes,
            },
          },
        });
      console.log("ISSUER CREDENTIAL");
      console.log(JSON.stringify(credentialExchangeRecord, null, 2));
      console.log("issue credential end");

      const response = new IssueCredentialResponseDto();
      response.credentialId = credentialExchangeRecord.id;
      response.connectionId = credentialExchangeRecord.connectionId;
      response.attributes = credentialExchangeRecord.credentialAttributes;
      response.createdAt = credentialExchangeRecord.createdAt;

      return response;
    } catch (e) {
      console.log("------------------------");
      console.log(JSON.stringify(e, null, 2));
      console.log("------------------------");
      throw new Error("unknown");
    }
  };

  acceptCredential = async (credentialRecordId: string) => {
    const linkSecretIds =
      await this.askar.agent.modules.anoncreds.getLinkSecretIds();

    if (linkSecretIds.length === 0) {
      await this.askar.agent.modules.anoncreds.createLinkSecret();
    }

    const credentialExchangeRecord =
      await this.askar.agent.credentials.acceptOffer({
        credentialRecordId,
      });

    const response = new IssueCredentialResponseDto();
    response.credentialId = credentialExchangeRecord.id;
    response.connectionId = credentialExchangeRecord.connectionId;
    response.attributes = credentialExchangeRecord.credentialAttributes;
    response.createdAt = credentialExchangeRecord.createdAt;

    return response;
  };

  credentials = async () => {
    const credentials = await this.askar.agent.credentials.findAllByQuery({
      state: CredentialState.Done,
    });

    const response: Array<IssueCredentialResponseDto> = [];
    for (const offer of credentials) {
      const t = new IssueCredentialResponseDto();
      t.credentialId = offer.id;
      t.connectionId = offer.connectionId;
      t.createdAt = offer.createdAt;
      t.attributes = offer.credentialAttributes;
      response.push(t);
    }

    return response;
  };

  credentialByStatedOfferReceived = async () => {
    const offers = await this.askar.agent.credentials.findAllByQuery({
      state: CredentialState.OfferReceived,
    });

    const response: Array<IssueCredentialResponseDto> = [];
    for (const offer of offers) {
      const t = new IssueCredentialResponseDto();
      t.credentialId = offer.id;
      t.connectionId = offer.connectionId;
      t.createdAt = offer.createdAt;
      t.attributes = offer.credentialAttributes;

      response.push(t);
    }

    return response;
  };

  issueProof = async (issueProofDto: IssueProofRequestDto) => {
    let exchangeRecord: ProofExchangeRecord;
    const requestedAttributes: Record<string, AnonCredsRequestedAttribute> = {};

    for (const attr of issueProofDto.attributes) {
      requestedAttributes[attr.attributeName] = {
        name: attr.attributeName,
        restrictions: [
          {
            schema_id: attr.schemaId,
            cred_def_id: attr.credentialDefinitionId,
          },
        ],
      };
    }

    if (!issueProofDto.connectionId) {
      const { proofRecord } = await this.askar.agent.proofs.createRequest({
        protocolVersion: "v2",
        proofFormats: {
          anoncreds: {
            name: "proof-request",
            version: "1.0",
            requested_attributes: requestedAttributes,
          },
        },
      });

      exchangeRecord = proofRecord;
    } else {
      exchangeRecord = await this.askar.agent.proofs.requestProof({
        protocolVersion: "v2",
        connectionId: issueProofDto.connectionId,
        proofFormats: {
          anoncreds: {
            name: "proof-request",
            version: "1.0",
            requested_attributes: requestedAttributes,
          },
        },
      });
    }

    const response = new IssueProofResponseDto();
    response.proofId = exchangeRecord.id;
    response.connectionId = exchangeRecord.connectionId;
    response.state = exchangeRecord.state;
    response.updatedAt = exchangeRecord.updatedAt;
    response.createdAt = exchangeRecord.createdAt;

    return response;
  };

  proofs = async () => {
    const proofs = await this.askar.agent.proofs.findAllByQuery({
      state: ProofState.RequestReceived,
    });

    const response: Array<IssueProofResponseDto> = [];
    for (const proof of proofs) {
      const t = new IssueProofResponseDto();
      t.proofId = proof.id;
      t.connectionId = proof.connectionId;
      t.state = proof.state;
      t.updatedAt = proof.updatedAt;
      t.createdAt = proof.createdAt;

      response.push(t);
    }

    return response;
  };

  acceptProof = async (proofRecordId: string) => {
    const requestedCredentials =
      await this.askar.agent.proofs.selectCredentialsForRequest({
        proofRecordId,
      });

    const proof = await this.askar.agent.proofs.acceptRequest({
      proofRecordId,
      proofFormats: requestedCredentials.proofFormats,
    });

    const t = new IssueProofResponseDto();
    t.proofId = proof.id;
    t.connectionId = proof.connectionId;
    t.state = proof.state;
    t.updatedAt = proof.updatedAt;
    t.createdAt = proof.createdAt;

    return t;
  };

  resolve = async (did: string) => {
    return this.askar.agent.dids.resolve(did);
  };

  sendMessage = async (message: MakeBasicMessageRequestDto) => {
    const response = new MakeBasicMessageResponseDto();

    const m = await this.askar.agent.basicMessages.sendMessage(
      message.connectionId,
      message.message,
    );

    const connectionInfo = await this.getConnectionById(message.connectionId);

    response.connectionId = m.connectionId;
    response.to = connectionInfo?.theirLabel || "";
    response.id = m.id;
    response.message = m.content;

    return response;
  };
}
